#!/bin/bash
# Copyright (c) 2018 Zilliqa
# This source code is being disclosed to you solely for the purpose of your
# participation in testing Zilliqa. You may view, compile and run the code for
# that purpose and pursuant to the protocols and algorithms that are programmed
# into, and intended by, the code. You may not do anything else with the code
# without express permission from Zilliqa Research Pte. Ltd., including
# modifying or publishing the code (or any part of it), and developing or
# forming another public or private blockchain network. This source code is
# provided 'as is' and no warranties are given as to title or non-infringement,
# merchantability or fitness for purpose and, to the extent permitted by law,
# all liability for your use of the code is disclaimed. Some programs in this
# code are governed by the GNU General Public License v3.0 (available at
# https://www.gnu.org/licenses/gpl-3.0.en.html) ('GPLv3'). The programs that
# are governed by GPLv3.0 are those programs that are located in the folders
# src/depends and tests/depends and which include a reference to GPLv3 in their
# program files.

#######################################################################################
# This script should be only used by Zilliqa Research,                                #
# for releasing a draft version binary with relative version information onto GitHub. #
#######################################################################################

# [MUST BE FILLED IN] User configuration settings
GitHubToken=""
packageName=""
releaseTitle=""
releaseDescription=""


# Environment variables
releaseDir="release"
versionFile="VERSION"
dsNodeFile="dsnodes.xml"
majorLine=2
minorLine=4
fixLine=6
DSLine=8
commitLine=12
shaLine=14
sigLine=16

# Validate input argument
if [ "$#" -ne 5 ]; then
    echo "Usage: source ./release.sh privateKeyFileName publicKeyFileName constantFileName constantLookupFileName constantArchivalFileName"
    return 1
fi

if [ ! -f "$1" ]; then
    echo "*ERROR* File : $1 not found!"
    return 1
fi

if [ ! -f "$2" ]; then
    echo "*ERROR* File : $2 not found!"
    return 1
fi

if [ ! -f "$3" ]; then
    echo "*ERROR* File : $3 not found!"
    return 1
fi

if [ ! -f "$4" ]; then
    echo "*ERROR* File : $4 not found!"
    return 1
fi

if [ ! -f "$5" ]; then
    echo "*ERROR* File : $5 not found!"
    return 1
fi

if [ "$GitHubToken" = "" ]; then
    echo "*ERROR* Please enter your own GitHub token in release.sh!"
    return 1
fi

if [ "$packageName" = "" ]; then
    echo "*ERROR* Please enter the package name in release.sh!"
    return 1
fi

if [ "$releaseTitle" = "" ] || [ "$releaseDescription" = "" ]; then
    echo "*ERROR* Please enter the release title and description in release.sh!"
    return 1
fi

# Read information from files
constantFile="$(realpath $3)"
constantLookupFile="$(realpath $4)"
constantArchivalFile="$(realpath $5)"
accountName="$(grep -oPm1 "(?<=<UPGRADE_HOST_ACCOUNT>)[^<]+" ${constantFile})"
repoName="$(grep -oPm1 "(?<=<UPGRADE_HOST_REPO>)[^<]+" ${constantFile})"
defaultMajor="$(sed -n ${majorLine}p ${versionFile})"
defaultMinor="$(sed -n ${minorLine}p ${versionFile})"
defaultFix="$(sed -n ${fixLine}p ${versionFile})"
defaultDS="$(sed -n ${DSLine}p ${versionFile})"
defaultCommit="$(sed -n ${commitLine}p ${versionFile})"
currentVer=${defaultMajor}.${defaultMinor}.${defaultFix}.${defaultDS}.${defaultCommit}

# Ask user to input new version information
echo -e "Current software version:  ${currentVer}"
read -p "Please enter major version [${defaultMajor}]: " major
major=${major:-${defaultMajor}}
read -p "Please enter minor version [${defaultMinor}]: " minor
minor=${minor:-${defaultMinor}}
read -p "Please enter fix version [${defaultFix}]: " fix
fix=${fix:-${defaultFix}}
read -p "Please enter expected DS epoch [${defaultDS}]: " DS
DS=${DS:-${defaultDS}}
commit="$(git describe --always)"

# Write new version information into version file
echo -e "Writing new software version into ${versionFile}..."
sed -i "${majorLine}s/.*/${major}/" ${versionFile}
sed -i "${minorLine}s/.*/${minor}/" ${versionFile}
sed -i "${fixLine}s/.*/${fix}/" ${versionFile}
sed -i "${DSLine}s/.*/${DS}/" ${versionFile}
sed -i "${commitLine}s/.*/${commit}/" ${versionFile}
newVer=${major}.${minor}.${fix}.${DS}.${commit}
export ZIL_VER=${newVer}
export ZIL_PACK_NAME=${packageName}
echo -e "New software version: ${newVer} is written into ${versionFile} successfully.\n"

# Use cpack to making deb file
echo -e "Make deb package..."
rm -rf ${releaseDir}
cmake -H. -B${releaseDir} -DCMAKE_BUILD_TYPE=RelWithDebInfo -DCMAKE_INSTALL_PREFIX=/usr/local/
cmake --build ${releaseDir} --j4
cd ${releaseDir}; make package; cp ../${versionFile} .; debFile="$(ls *.deb)"; cd -
echo -e "Deb packages are generated successfully.\n"

# Make SHA-256 & multi-signature
echo -e "Making SHA-256 & multi-signature..."
privKeyFile="$(realpath $1)"
pubKeyFile="$(realpath $2)"
cd ${releaseDir}
sha="$(sha256sum ${debFile}|cut -d ' ' -f1|tr 'a-z' 'A-Z')"
sed -i "${shaLine}s/.*/${sha}/" ${versionFile}
signature="$(./bin/signmultisig ${sha} ${privKeyFile} ${pubKeyFile})"
sed -i "${sigLine}s/.*/${signature}/" ${versionFile}
cd -
echo -e "SHA-256 & multi-signature are written into ${versionFile} successfully.\n"

#Sign the ds nodes and update the xml
cd ${releaseDir}
echo -e "Updating ds xml file"
cp ../constants_local.xml ./constants.xml
cp ../${dsNodeFile} ./${dsNodeFile}
ret="$(./bin/gensigninitialds   ${privKeyFile} ${pubKeyFile})"
cp ./${dsNodeFile} ../${dsNodeFile}
cd - 

# Upload package onto GitHub
echo -e "Creating new release and uploading package onto GitHub..."
fullCommit="$(git rev-parse HEAD)"
releaseLog="release.log"
curl -v -s \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/json" "https://api.github.com/repos/${accountName}/${repoName}/releases" \
  -d '{
  "tag_name": "'"${newVer}"'", 
  "target_commitish": "'"${fullCommit}"'",
  "name": "'"${releaseTitle}"'",
  "body": "'"${releaseDescription}"'",
  "draft": true,
  "prerelease": false
}' > ${releaseLog}

line="$(sed '6!d' ${releaseLog})"
releaseId=${line:8:8}
check='^[0-9]+$'
if ! [[ $releaseId =~ $check ]] ; then
    echo -e "*ERROR* Create new release fail! Please check input value and ${releaseLog}, then try again."
    return 1
fi
curl -v -s \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/octet-stream" \
  --data-binary @${pubKeyFile} \
  "https://uploads.github.com/repos/${accountName}/${repoName}/releases/${releaseId}/assets?name=$(basename {pubKeyFile})"
curl -v -s \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/vnd.debian.binary-package" \
  --data-binary @${releaseDir}/${debFile} \
  "https://uploads.github.com/repos/${accountName}/${repoName}/releases/${releaseId}/assets?name=${debFile}"
curl -v -s \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/octet-stream" \
  --data-binary @${releaseDir}/${versionFile} \
  "https://uploads.github.com/repos/${accountName}/${repoName}/releases/${releaseId}/assets?name=${versionFile}"
curl -v -s  \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/octet-stream"  \
  --data-binary @"${dsNodeFile}" \
  "https://uploads.github.com/repos/${accountName}/${repoName}/releases/${releaseId}/assets?name=${dsNodeFile}"
curl -v -s  \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/octet-stream"  \
  --data-binary @"${constantFile}" \
  "https://uploads.github.com/repos/${accountName}/${repoName}/releases/${releaseId}/assets?name=${constantFile##*/}"
curl -v -s  \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/octet-stream"  \
  --data-binary @"${constantLookupFile}" \
  "https://uploads.github.com/repos/${accountName}/${repoName}/releases/${releaseId}/assets?name=${constantLookupFile##*/}_lookup"
curl -v -s  \
  -H "Authorization: token ${GitHubToken}" \
  -H "Content-Type:application/octet-stream"  \
  --data-binary @"${constantArchivalFile}" \
  "https://uploads.github.com/repos/${accountName}/${repoName}/releases/${releaseId}/assets?name=${constantArchivalFile##*/}_archival"
rm ${releaseLog}
echo -e "\nA new draft release with package is created on Github sucessfully, please proceed to publishing the draft release on Github webpage.\n"
